// I2C/TWI Code. Basierte ursprnglich auf Atmel Application Note AVR312 "Using the USI module as a I2C slave" sowie Atmel Application Note AVR310 "Using the USI as a I2C master"
// angepasst fr den ATtiny84 mit kleinen nderungen und Ergnzungen

#include "main.h"


#ifdef USE_TWI_MASTER
/************************************************************
   Funktionsprototypen und Variablen fr den I2C Master
************************************************************/
unsigned char I2CMasterTransfer(unsigned char);
unsigned char I2CMasterStop(void);

union USIMasterState
{
	unsigned char errorState;	// can reuse the TWI_state for error states due to that it will not be need if there exists an error
	struct
	{
		unsigned char addressMode : 1;
		unsigned char masterWriteDataMode : 1;
		unsigned char unused : 6;
	}; 
} USIMasterState;
#endif


#ifdef USE_TWI_SLAVE
/************************************************************
      Funktionen und Variablen fr den I2C Slave
************************************************************/
volatile unsigned char USIOverflowState;	// state variable used for the USI overflow interrupt

// ringbuffers
unsigned char I2CReceiveBuffer[I2C_RX_BUFFER_SIZE];
volatile unsigned char I2CReceiveBufferHead;
volatile unsigned char I2CReceiveBufferTail;

unsigned char I2CTransmitBuffer[I2C_TX_BUFFER_SIZE];
volatile unsigned char I2CTransmitBufferHead;
volatile unsigned char I2CTransmitBufferTail;


// flushes the I2C ringbuffers
void I2CSlaveFlushBuffers(void)
{
	I2CReceiveBufferTail = 0;
	I2CReceiveBufferHead = 0;
	I2CTransmitBufferTail = 0;
	I2CTransmitBufferHead = 0;
}


// initializes USI for TWI slave mode
void I2CSlaveInit(void)
{
	I2CSlaveFlushBuffers();
	USICR = (1 << USIWM1) | (1 << USICS1);
	USISR = 0xF0;	// clear all flags and reset overflow counter
	DDR_USI &= ~(1<<PORT_USI_SDA);	// set SDA as input
	PORT_USI &= ~(1<<PORT_USI_SDA);	// set SDA high
	DDR_USI |= (1<<PORT_USI_SCL);			// set SCL as output
	PORT_USI |= (1<<PORT_USI_SCL);		// set SCL high
	USICR |= (1 << USISIE);
}


// puts data in the transmission buffer. Waits if buffer is full
void I2CSlaveTransmitByte(unsigned char data)
{
    unsigned char tmphead = (I2CTransmitBufferHead + 1) & I2C_TX_BUFFER_MASK;	// calculate buffer index
    while (tmphead == I2CTransmitBufferTail) ;	// wait for free space in buffer
    I2CTransmitBuffer[tmphead] = data;	// store data in buffer
    I2CTransmitBufferHead = tmphead;	// store new index
}


// returns a byte from the receive buffer. Waits if buffer is empty
unsigned char I2CSlaveReceiveByte(void)
{
    while (I2CReceiveBufferHead == I2CReceiveBufferTail) ;
	unsigned char tmptail = (I2CReceiveBufferTail + 1) & I2C_RX_BUFFER_MASK;	// calculate buffer index
    I2CReceiveBufferTail = tmptail;	// store new index
    return I2CReceiveBuffer[tmptail];	// return data from the buffer
}


// check if there is data in the receive buffer
unsigned char I2CSlaveDataInReceiveBuffer(void)
{
    return (I2CReceiveBufferHead != I2CReceiveBufferTail);	// return 0 (FALSE) if the receive buffer is empty
}


// USI start condition ISR. Detects the USI_TWI Start Condition and intialises the USI for reception of the "TWI Address" packet
ISR(USI_STR_vect)
{
	// set default starting conditions for new TWI package
    USIOverflowState = USI_SLAVE_CHECK_ADDRESS;
    DDR_USI  &= ~(1<<PORT_USI_SDA);	// SDA as input
	while (PIN_USI & (1 << PORT_USI_SCL)) ;	// wait for SCL to go low to ensure the "Start Condition" has completed
	USISR = (1 << USISIF) | (1 << USIOIF) | (1 << USIPF) | (1 << USIDC); // clear flags
	USISR = 0xF0;
	USICR |= (1 << USIOIE) | (1 << USIWM0);	// enable overflow interrupt and clk stretching
}


// USI counter overflow ISR. Handles all the comunication. Is disabled only when waiting for new start condition
ISR(USI_OVF_vect)
{
	unsigned char tmpTxTail;
	unsigned char tmpUSIDR = USIDR;
	
	switch (USIOverflowState)
	{
		// check slave address
		case USI_SLAVE_CHECK_ADDRESS:
			if ((tmpUSIDR >> 1) == configuration.I2CSlaveAddress)	// check slave address
			{
				DDR_USI |=  (1<<PORT_USI_SDA); // SDA as output
				USISR = (1 << USICNT3) | (1 << USICNT2) | (1 << USICNT1);	// reload counter value for ACK
				
				if (tmpUSIDR & 0x01)	// check if read or write mode
					USIOverflowState = USI_SLAVE_SEND_DATA;
				else
					USIOverflowState = USI_SLAVE_REQUEST_DATA;
			}
			else	// not my address
				USICR &= ~((1 << USIOIE) | (1 << USIWM0));	// set start condition detection mode
		break;

		// master read data mode
		// check reply and goto USI_SLAVE_SEND_DATA if OK, else reset USI.
		case USI_SLAVE_CHECK_REPLY_FROM_SEND_DATA:
			if (tmpUSIDR & 1) // if NACK, the master does not want more data
			{
				USIOverflowState = USI_SLAVE_CHECK_ADDRESS;
				USICR &= ~((1 << USIOIE) | (1 << USIWM0));	// set start condition detection mode
				break;
			}
			else
				DDR_USI |= (1 << PORT_USI_SDA);	// SDA as output

		// copy data from buffer to USIDR and set USI to shift byte. Next USI_SLAVE_REQUEST_REPLY_FROM_SEND_DATA
		case USI_SLAVE_SEND_DATA:
			// get data from ringbuffer
			tmpTxTail = I2CTransmitBufferTail;
			if (I2CTransmitBufferHead != tmpTxTail)
			{
				I2CTransmitBufferTail = (I2CTransmitBufferTail + 1) & I2C_TX_BUFFER_MASK;
				USIDR = I2CTransmitBuffer[I2CTransmitBufferTail];
			}
			else // if the buffer is empty then, send zero
				USIDR = 0;
			
			PORT_USI |= (1 << PORT_USI_SDA);	// SDA high
			USIOverflowState = USI_SLAVE_REQUEST_REPLY_FROM_SEND_DATA;
		break;

		// set USI to sample reply from master. Next USI_SLAVE_CHECK_REPLY_FROM_SEND_DATA
		case USI_SLAVE_REQUEST_REPLY_FROM_SEND_DATA:
			DDR_USI &= ~(1 << PORT_USI_SDA);	// SDA as input
			PORT_USI &= ~(1 << PORT_USI_SDA); // SDA low
			USISR = (1 << USICNT3) | (1 << USICNT2) | (1 << USICNT1);	// reload counter value for ACK
			USIOverflowState = USI_SLAVE_CHECK_REPLY_FROM_SEND_DATA;
		break;

		// master write data mode
		// set USI to sample data from master. Next USI_SLAVE_GET_DATA_AND_SEND_ACK
		case USI_SLAVE_REQUEST_DATA:
			DDR_USI &= ~(1 << PORT_USI_SDA);	// SDA as input
			USIOverflowState = USI_SLAVE_GET_DATA_AND_SEND_ACK;
		break;

		// copy data from USIDR and send ACK. Next USI_SLAVE_REQUEST_DATA
		case USI_SLAVE_GET_DATA_AND_SEND_ACK:
			// put data into ringbuffer
			I2CReceiveBufferHead = (I2CReceiveBufferHead + 1) & I2C_RX_BUFFER_MASK;
			I2CReceiveBuffer[I2CReceiveBufferHead] = tmpUSIDR;
			DDR_USI |=  (1<<PORT_USI_SDA); // SDA as output
			USISR = (1 << USICNT3) | (1 << USICNT2) | (1 << USICNT1);	// reload counter value for ACK
			USIOverflowState = USI_SLAVE_REQUEST_DATA;
		break;
	}
	USISR |= (1 << USIOIF);	// clear overflow flag, releases SCL
}
#endif


#ifdef USE_TWI_MASTER
/************************************************************
                 Funktionen fr den I2C Master
************************************************************/

// USI TWI single master initialization function
void I2CMasterInit(void)
{
	PORT_USI |= (1<<PIN_USI_SDA);	// enable pullup on SDA, to set high as released state
	PORT_USI |= (1<<PIN_USI_SCL);	// enable pullup on SCL, to set high as released state
	DDR_USI |= (1<<PIN_USI_SCL);	// enable SCL as output
	DDR_USI |= (1<<PIN_USI_SDA);	// enable SDA as output
	USIDR = 0xFF;	// preload dataregister with "released level" data
	USICR = (0<<USISIE) | (0<<USIOIE) | (1<<USIWM1) | (0<<USIWM0) | (1<<USICS1) | (0<<USICS0) | (1<<USICLK) | (0<<USITC); // disable Interrupts, set USI in Two-wire mode, software stobe as counter clock source
	USISR = (1<<USISIF) | (1<<USIOIF) | (1<<USIPF) | (1<<USIDC) | (0x0<<USICNT0);	// clear flags and reset counter
}


// returns the error message from the last transmission
unsigned char I2CMasterGetErrorState(void)
{
	return USIMasterState.errorState;	// return error state
}


// USI Transmit and receive function. The LSB of first byte in data indicates if a read or write cycles is performed. If set a read operation is performed.
// The function generates a (repeated) start condition, sends address and the read/write bit, reads/writes data and verifies/sends ACK.
// Success or error code is returned. The error codes are defined in twi.h
unsigned char I2CMasterStartTransfer(unsigned char *msg, unsigned char msgSize)
{
	unsigned char tempUSISR_8bit = (1<<USISIF) | (1<<USIOIF) | (1<<USIPF) | (1<<USIDC) | (0x0<<USICNT0);	// clear flags and set USI to shift 8 bits i.e. count 16 clock edges
	unsigned char tempUSISR_1bit = (1<<USISIF) | (1<<USIOIF) | (1<<USIPF) | (1<<USIDC) | (0xE<<USICNT0);	// clear flags and set USI to shift 1 bit i.e. count 2 clock edges
	
	USIMasterState.errorState = 0;
	USIMasterState.addressMode = TRUE;
	
	if (!(*msg & 1))	// the LSB in the address byte determines if it is a masterRead or masterWrite operation
		USIMasterState.masterWriteDataMode = TRUE;

	// release SCL to ensure that (repeated) start can be performed
	PORT_USI |= (1<<PIN_USI_SCL);	// release SCL
	while (!(PORT_USI & (1<<PIN_USI_SCL))) ;	// verify that SCL becomes high
	_delay_us(TWI_DELAY_T4);	// delay for T4TWI
	
	// generate Start Condition
	PORT_USI &= ~(1<<PIN_USI_SDA);	// force SDA low
	_delay_us(TWI_DELAY_T4);
	PORT_USI &= ~(1<<PIN_USI_SCL);	// pull SCL low
	PORT_USI |= (1<<PIN_USI_SDA);		// release SDA

	// write address and read/write data
	do
	{
		// if masterWrite cycle (or inital address tranmission)
		if (USIMasterState.addressMode || USIMasterState.masterWriteDataMode)
		{
			// write a byte
			PORT_USI &= ~(1<<PIN_USI_SCL); // pull SCL low
			USIDR = *(msg++);	// setup data
			I2CMasterTransfer(tempUSISR_8bit);	// send 8 bits on bus
			
			// clock and verify (N)ACK from slave
			DDR_USI  &= ~(1<<PIN_USI_SDA);	// enable SDA as input
			if (I2CMasterTransfer(tempUSISR_1bit) & 1)
			{
				if (USIMasterState.addressMode)
					USIMasterState.errorState = USI_TWI_NO_ACK_ON_ADDRESS;
				else
					USIMasterState.errorState = USI_TWI_NO_ACK_ON_DATA;
				return (FALSE);
			}
			USIMasterState.addressMode = FALSE;	// only perform address transmission once
		}
		// else masterRead cycle
		else
		{
			// read a data byte
			DDR_USI &= ~(1<<PIN_USI_SDA);	// enable SDA as input
			*(msg++) = I2CMasterTransfer(tempUSISR_8bit);
			
			// prepare to generate ACK (or NACK in case of end of transmission)
			if (msgSize == 1)	// if transmission of last byte was performed
				USIDR = 0xFF;	// load NACK to confirm end of transmission
			else
				USIDR = 0x00;	// load ACK, set data register bit 7 (output for SDA) low
			I2CMasterTransfer(tempUSISR_1bit);	// generate ACK/NACK
		}
	} while (--msgSize);	// until all data sent/received
	
	I2CMasterStop();	// send a STOP condition on the TWI bus
	return TRUE;	// transmission successfully completed
}


// core function for shifting data in and out from the USI. The data to be sent has to be placed into the USIDR prior to calling this function. Data read will be returned from the function
unsigned char I2CMasterTransfer(unsigned char temp)
{
	USISR = temp;	// set USISR according to temp
	// prepare clocking
	temp = (0<<USISIE) | (0<<USIOIE) | (1<<USIWM1) | (0<<USIWM0) | (1<<USICS1) | (0<<USICS0) | (1<<USICLK) | (1<<USITC);	// interrupts disabled, set USI in Two-wire mode, software clock strobe as source, toggle clock port
	do
	{
		_delay_us(TWI_DELAY_T2);
		USICR = temp;	// generate positve SCL edge
		while (!(PIN_USI & (1<<PIN_USI_SCL))) ;	// wait for SCL to go high
		_delay_us(TWI_DELAY_T4);
		USICR = temp;	// generate negative SCL edge
	} while (!(USISR & (1<<USIOIF)));	// check for transfer complete
	
	_delay_us(TWI_DELAY_T2);
	temp  = USIDR;	// read out data
	USIDR = 0xFF;	// release SDA
	DDR_USI |= (1<<PIN_USI_SDA);	// enable SDA as output

	return temp;	// return the data from the USIDR
}


// generates a TWI stop condition and releases the bus
unsigned char I2CMasterStop(void)
{
	PORT_USI &= ~(1<<PIN_USI_SDA);	// pull SDA low
	PORT_USI |= (1<<PIN_USI_SCL);		// release SCL
	while (!(PIN_USI & (1<<PIN_USI_SCL))) ;	// wait for SCL to go high
	_delay_us(TWI_DELAY_T4);
	PORT_USI |= (1<<PIN_USI_SDA);		// release SDA
	_delay_us(TWI_DELAY_T2);
	return TRUE;
}
#endif

