// basiert auf Atmel Application Note AVR307: Half Duplex UART Using the USI Module
// angepasst fr den ATtiny84 mit kleinen nderungen und Ergnzungen

#include "main.h"

#ifdef USE_UART

register unsigned char USI_UART_TxData asm("r2");

// UART ringbuffer
unsigned char UART_RxBuf[UART_RX_BUFFER_SIZE];
volatile unsigned char UART_RxHead;
volatile unsigned char UART_RxTail;
unsigned char UART_TxBuf[UART_TX_BUFFER_SIZE];
volatile unsigned char UART_TxHead;
volatile unsigned char UART_TxTail;

volatile union USI_UART_status	// Status byte holding flags.
{
	unsigned char status;
	struct
	{
		unsigned char ongoingTransmissionFromBuffer : 1;
		unsigned char ongoingTransmissionOfPackage : 1;
		unsigned char ongoingReceptionOfPackage : 1;
		unsigned char receptionBufferOverflow : 1;
		unsigned char flag4 : 1;
		unsigned char flag5 : 1;
		unsigned char flag6 : 1;
		unsigned char flag7 : 1;
	};
} USI_UART_status = {0};



void initUART(void)
{
	DDRA &= ~((1 << PA5) | (1 << PA6));	// set RX and TX as input
	PORTA |= (1 << PA5) | (1 << PA6);	// enable pullup on RX and set TX high
	PCMSK0 |= (1 << PCINT6);	// enable pin change for PA6
	TCCR0A = (0 << COM0A1) | (0 << COM0A0) | (0 << COM0B1) | (0 << COM0B0) | (1 << WGM01) | (0 << WGM00);	// CTC mode
	uartFlushBuffers();
	uartInitializeReceiver();
}


// reverses the order of bits in a byte (MSB is swapped with LSB, etc.)
unsigned char bitReverse(unsigned char x)
{
	x = ((x >> 1) & 0x55) | ((x << 1) & 0xaa);
	x = ((x >> 2) & 0x33) | ((x << 2) & 0xcc);
	x = ((x >> 4) & 0x0f) | ((x << 4) & 0xf0);
	return x;    
}


// flushes the UART buffers
void uartFlushBuffers(void)
{
	UART_RxTail = 0;
	UART_RxHead = 0;
	UART_TxTail = 0;
	UART_TxHead = 0;
}


// initializes USI for UART transmission
void uartInitializeTransmitter(void)
{
	cli();	// disable interrupt
	TCNT0 = 0;
	OCR0A = TIMER0_SEED;
	GTCCR |= (1 << PSR10);	// reset prescaler
	TCCR0A = (0 << COM0A1) | (0 << COM0A0) | (0 << COM0B1) | (0 << COM0B0) | (1 << WGM01) | (0 << WGM00);	// CTC mode
	TCCR0B = (0 << FOC0A) | (0 << FOC0B) | (0 << WGM02) | (0 << CS02) | (1 << CS01) | (0 << CS00); // clock = 1 MHz
	TIFR0 |= (1 << OCIE0A);	// clear timer interrupt flag

	USICR = (0<<USISIE) | (1<<USIOIE) | (0<<USIWM1) | (1<<USIWM0) | (0<<USICS1) | (1<<USICS0) | (0<<USICLK) | (0<<USITC);	// USI overflow interrupt, three wire mode, timer0 as clock source
	USIDR = 0xFF;	// make sure MSB is '1' before enabling USI_DO.
	USISR = 0xFF;	// clear all USI interrupt flags and preload the USI counter to generate interrupt at first USI clock
	
	GIMSK &= ~(1<<PCIE0);	// disable pin change interrupt for PA6
	GIFR |= (1<<PCIF0);		// clear pin change interrupt flag
	DISABLE_TIMER1_INTERRUPT();
	DDRA |= (1 << PA5);
	
	USI_UART_status.ongoingTransmissionFromBuffer = TRUE;
	sei();	// re-enable interrupts
}


// initializes USI for UART reception
void uartInitializeReceiver(void)
{
	PORTA |= (1 << PA5);
	DDRA &= ~(1 << PA5);
	USICR = 0;	// disable USI
	GIFR |= (1<<PCIF0);	// clear pin change interrupt flag
	GIMSK |= (1<<PCIE0);	// enable pin change interrupt
}


// puts data in the transmission buffer, after reverseing the bits in the byte
// and initiates the transmission routines if not already started
void uartTransmitByte(unsigned char data)
{
	unsigned char tmphead = (UART_TxHead + 1) & UART_TX_BUFFER_MASK;	// calculate buffer index
	while (tmphead == UART_TxTail)
		;	// wait for free space in buffer
	UART_TxBuf[tmphead] = bitReverse(data);	// reverse the order of the bits in the data byte and store data in buffer
	UART_TxHead = tmphead;	// store new index
	
	if (!USI_UART_status.ongoingTransmissionFromBuffer)	// start transmission from buffer (if not already started)
	{
		while (USI_UART_status.ongoingReceptionOfPackage)
			;	// wait for USI to finsh reading incoming data.
		uartInitializeTransmitter();
	}
}


// Returns a byte from the receive buffer. Waits if buffer is empty.
unsigned char uartReceiveByte(void)
{
	unsigned char UART_RxTail_Temp = UART_RxTail;

	while (UART_RxHead == UART_RxTail_Temp)
		;	// wait for incoming data 
	unsigned char tmptail = (UART_RxTail_Temp + 1) & UART_RX_BUFFER_MASK;	// calculate buffer index 
	UART_RxTail = tmptail; // store new index 
	return bitReverse(UART_RxBuf[tmptail]);	// reverse the order of the bits in the data byte before it returns data from the buffer
}


// pin change interrupt on RX
// USI is started to sample the RX line
ISR(PCINT0_vect)
{
	if (!(PINA & (1<<PA6)))	// if the USI DI pin is low
	{
		DISABLE_TIMER1_INTERRUPT();
		OCR0A = INITIAL_TIMER0_SEED;   // set compare value to generate compare match on first byte after start bytes (1,5 bit times from falling edge of start bit)
		GTCCR |= (1 << PSR10);	// reset prescaler
		TCCR0A = (0 << COM0A1) | (0 << COM0A0) | (0 << COM0B1) | (0 << COM0B0) | (1 << WGM01) | (0 << WGM00);	// CTC mode
		TCCR0B = (0 << FOC0A) | (0 << FOC0B) | (0 << WGM02) | (0 << CS02) | (1 << CS01) | (0 << CS00);	// start timer, clock = 1 MHz
		TCNT0 = INTERRUPT_STARTUP_DELAY;	// set timer value to interrupt latency
		TIFR0 |= (1 << OCIE0A);	// reset timer interrupt flag
		TIMSK0 |= (1 << OCIE0A);	// enable timer compare match
		
		USICR = (0<<USISIE) | (1<<USIOIE) | (0<<USIWM1) | (1<<USIWM0) | (0<<USICS1) | (1<<USICS0) | (0<<USICLK) | (0<<USITC);	// USI overflow interrupt, three wire mode, timer0 as clock source
		USISR = 0xF0 | USI_COUNTER_SEED_RECEIVE;	// clear all USI interrupt flags and preload the USI counter to generate interrupt
		GIMSK &= ~(1<<PCIE0);	// disable pin change interrupt for PA6
		USI_UART_status.ongoingReceptionOfPackage = TRUE;
		if (configuration.ppmEnable > 1)
			configuration.ppmEnable = 2;
	}
	else if ((PINA & (1 << PA4)) == 0)
	{
		ppmLow = TCNT0;
		ppmOk = 1;
	}
	else
		ppmHigh = TCNT0;
}


// timer 0 compare match
ISR(TIM0_COMPA_vect)
{
	OCR0A = TIMER0_SEED;	// set compare value according to baud rate
	TIMSK0 &= ~(1 << OCIE0A);	// disable compare match interrupt
}


// USI Counter Overflow interrupt
// it is used for moving data between memory and the USI data register (is used for both transmission and reception)
ISR(USI_OVF_vect)
{
	// check if we are running in transmit mode
	union USI_UART_status USI_UART_status_Temp = USI_UART_status;

	if (USI_UART_status_Temp.ongoingTransmissionFromBuffer)
	{
		// if ongoing transmission, then send second half of transmit data
		if (USI_UART_status_Temp.ongoingTransmissionOfPackage)
		{                                   
			USI_UART_status_Temp.ongoingTransmissionOfPackage = FALSE;	// clear on-going package transmission flag
			USISR = 0xF0 | (USI_COUNTER_SEED_TRANSMIT);	// load USI Counter seed and clear all USI flags
			USIDR = (USI_UART_TxData << 3) | 0x07;	// reload the USIDR with the rest of the data and a stop-bit
		}
		else	// else start sending more data or leave transmit mode
		{
			// if there is data in the transmit buffer, then send first half of data
			unsigned char UART_TxTail_Temp = UART_TxTail;

			if (UART_TxHead != UART_TxTail_Temp)
			{
				USI_UART_status_Temp.ongoingTransmissionOfPackage = TRUE;	// set on-going package transmission flag
				unsigned char tmptail = (UART_TxTail_Temp + 1) & UART_TX_BUFFER_MASK;	// calculate buffer index
				UART_TxTail = tmptail;	// store new index
				USI_UART_TxData = UART_TxBuf[tmptail];	// read out the data that is to be sent. Note that the data must be bit reversed before sent
				// the bit reversing is moved to the application section to save time within the interrupt
				USISR = 0xF0 | (USI_COUNTER_SEED_TRANSMIT);	// load USI Counter seed and clear all USI flags
				USIDR = (USI_UART_TxData >> 2) | 0x80;	// copy (initial high state,) start-bit and 6 LSB of original data (6 MSB of bit of bit reversed data)
			}
			else	// else enter receive mode
			{
				USI_UART_status_Temp.ongoingTransmissionFromBuffer = FALSE;
				TCCR0B = 0;	// stop Timer0
				PORTA |= (1 << PA5);
				DDRA &= ~(1 << PA5);
				
				USICR = 0;	// disable USI
				GIFR |= (1<<PCIF0);	// clear pin change interrupt flag
				GIMSK |= (1<<PCIE0);	// enable pin change interrupt
				ENABLE_TIMER1_INTERRUPT();
				if (configuration.ppmEnable > 1)
					configuration.ppmEnable = 2;
			}
		}
	}
	else	// else running in receive mode
	{
		USI_UART_status_Temp.ongoingReceptionOfPackage = FALSE;           
		unsigned char tmphead = (UART_RxHead + 1) & UART_RX_BUFFER_MASK;	// calculate buffer index
		
		if (tmphead == UART_RxTail)	// if buffer is full trash data and set buffer full flag
			USI_UART_status_Temp.receptionBufferOverflow = TRUE;	// store status to take actions elsewhere in the application code
		else	// if there is space in the buffer then store the data.
		{
			UART_RxHead = tmphead;	// store new index
			UART_RxBuf[tmphead] = USIDR;	// store received data in buffer. Note that the data must be bit reversed before used
		}	// The bit reversing is moved to the application section to save time within the interrupt
		
		TCCR0B = 0;	// stop Timer0
		PORTA |= (1 << PA5);
		DDRA &= ~(1 << PA5);
		USICR = 0;	// disable USI
		GIFR |= (1<<PCIF0);	// clear pin change interrupt flag
		GIMSK |= (1<<PCIE0);	// enable pin change interrupt
		ENABLE_TIMER1_INTERRUPT();
		if (configuration.ppmEnable > 1)
			configuration.ppmEnable = 2;
	}
	USI_UART_status = USI_UART_status_Temp;
}


// sendet einen nullterminierten String, der im Flash liegt, via UART
inline void uartSendString_p(const char *progmemPtr)
{
    register char c;
    while ((c = pgm_read_byte(progmemPtr++)))
      uartTransmitByte(c);
}


// sendet einen nullterminierten String, der im RAM liegt, via UART
inline void uartSendString(const unsigned char *ptr)
{
    while (*ptr) 
      uartTransmitByte(*ptr++);
}


// sendet eine Zahl vom Typ unsigned long int per UART (ohne fhrende Nullen)
inline void uartSendNumber(unsigned long int number)
{
	uartTransmitByte(' ');
	unsigned long int rad = 1000000000UL;
	unsigned char zero = 1;
	while (rad != 0)
	{
		unsigned char r = number / rad % 10;
		if (r != 0 || zero == 0)
		{
			zero = 0;
			uartTransmitByte(r+48);
		}
		rad /= 10;
	}
	if (zero)
		uartTransmitByte('0');
}

#endif
